package complexthreads;

/**
 * Title:
 * Description:
 * Copyright:    Copyright (c) 2002
 * Company:
 * @author
 * @version 1.0
 */
import java.util.*;

class ManagedThread extends Thread
{
    int desiredState = ThreadStates.RUNNING;

    public int getDesiredState() {return desiredState;}

    public void mstop () {
        synchronized (this) {
            desiredState = ThreadStates.STOP;
            notifyAll();
        }
    }

    public void msuspend() {
		synchronized (this) {
			if (desiredState!=ThreadStates.STOP)
			{
				desiredState = ThreadStates.SLEEP;
			}
		}
	}

    public void mresume ()
    {
        synchronized (this) {
			if (desiredState != ThreadStates.STOP)	{
				desiredState = ThreadStates.RUNNING;
				notifyAll();
			}
        }
    }

    protected boolean keepRunning()
    {
        synchronized (this) {
            if (desiredState == ThreadStates.RUNNING)
                return true;
            else {
                while (true) {
                    if (desiredState == ThreadStates.STOP)
                    {
                        System.out.println(Thread.currentThread().getName()+ " stopped");
                        return false;
                    }
                    else if (desiredState == ThreadStates.SLEEP)
                        try {
                            System.out.println(Thread.currentThread().getName()+ " fell asleep");
                            this.wait();
                        } catch (Exception ex) {
                            System.err.println(Thread.currentThread().getName()+" interrupted");
                            return false;
                        }
                    else if (desiredState == ThreadStates.RUNNING) {
                        System.out.println(Thread.currentThread().getName()+" resumed");
                        return true;
                    }
                }
            }
        }
    }
}
class Consumer extends ManagedThread
{
    Storage s;
    long pause;

    Consumer (Storage s, long pause)
    {
        super();
        this.s = s;
        this.pause = pause;
    }
    public void run ()
    {
        while (true)
        {
            try {
                Thread.currentThread().sleep(pause);
                Object o = s.get();
                System.out.println(Thread.currentThread().getName()+ " got detail "+o);
                if (!keepRunning())
                    break;
            } catch (InterruptedException ex) {
                System.out.println(Thread.currentThread().getName()+ " interrputed");
                break;
            }
        }
    }
}
class Producer extends ManagedThread
{
    Storage s;
    long pause;

    Producer (Storage s, long pause)
    {
        super();
        this.s = s;
        this.pause = pause;
    }
    public void run ()
    {
        int number=0;
        while (true)
        {
            try {
                Thread.currentThread().sleep(pause);
                System.out.println(Thread.currentThread().getName()+ " put detail "+"Widget #"+number);
                s.put("Widget #"+number);
                number++;
                if (!keepRunning())
                   break;
            } catch (InterruptedException ex) {
                System.out.println(Thread.currentThread().getName()+" interrupted");
                break;
            }
        }
    }
}

class Storage {
    int i;
    LinkedList queue = new LinkedList();
    synchronized void put (Object o)
    {
        queue.add(o);
        notifyAll ();
    }
    synchronized Object get()
    {
        while (true)
        {
            if (!queue.isEmpty()) {
                return queue.remove(0);
            }
            else {
                try {
                    wait();
                } catch (InterruptedException ex) {
                    System.out.println("Interrupted in wait");
                    continue;
                }
            }
        }
    }
}

public class Main {
    public static void main(String[] args) throws InterruptedException {
        Main main1 = new Main();
        Storage s = new Storage();
        Consumer c = new Consumer (s,1000);
        Consumer c2 = new Consumer (s, 1100);
        Producer p = new Producer (s,600);
        c.setName("Consumer1");
        p.setName("Producer");
        c2.setName("Consumer2");
        c.start();
        c2.start();
        Thread.sleep(500);
        p.start();
        Thread.sleep(5000);
        p.msuspend();
        Thread.sleep(1000);
        c.interrupt();
        Thread.sleep(2100);
        p.mresume();
        Thread.sleep(2100);
        p.mstop();
        c.mstop();
        c2.mstop();
        p.join();
        c.join();
        c2.join();
        System.out.println("All threads finished");
    }
}
